"""
Analysis module for `pymead`. Contains a partially complete Python API to both XFOIL and the MSES software suite. Also
contains a linear strength vortex panel method for a fast, purely inviscid airfoil flow solution sped up using Numba.
"""